import typing

from . import _inspect
from .duration import Duration
from .indicators.Clef import Clef
from .indicators.StartBeam import StartBeam
from .indicators.StartHairpin import StartHairpin
from .indicators.StartTextSpan import StartTextSpan
from .indicators.StopBeam import StopBeam
from .indicators.StopHairpin import StopHairpin
from .indicators.StopTextSpan import StopTextSpan
from .instruments import Instrument
from .iterate import Iteration
from .iterpitches import sounding_pitches_are_in_range
from .parentage import Parentage
from .score import Container, Context
from .sequence import Sequence
from .tag import Tag

### PRIVATE FUNCTIONS ###


def _aggregate_context_wrappers(argument):
    """
    Special_Voice may contain other instances of Special_Voice.
    This currently happens with OnBeatGraceContainer.
    This method aggregates all Special_Voice wrappers for checks.
    """
    name_to_wrappers: typing.Dict = {}
    for context in Iteration(argument).components(Context):
        if context.name not in name_to_wrappers:
            name_to_wrappers[context.name] = []
        wrappers = context._dependent_wrappers[:]
        name_to_wrappers[context.name].extend(wrappers)
    return name_to_wrappers


### PUBLIC FUNCTIONS ###


def check_beamed_long_notes(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks beamed long notes.

    ..  container:: example

        Beamed quarter notes are not wellformed:

        >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'4
                [
                d'4
                ]
                e'4
                f'4
            }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        2 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

        Beamed eighth notes are wellformed:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'8
                [
                d'8
                ]
                e'8
                f'8
            }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

    ..  container:: example

        REGRESSION. Matching start- and stop-beam indicators work
        correctly:

        >>> voice = abjad.Voice("c'8 d'8 e'4 f'2")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'8
                [
                d'8
                ]
                e'4
                f'2
            }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

    The examples above feature Abjad voice containers because beams are
    voice-persistent.
    """
    violators, total = [], 0
    for leaf in Iteration(argument).leaves():
        total += 1
        if leaf.written_duration < Duration((1, 4)):
            continue
        start_wrapper = _inspect._get_effective(leaf, StartBeam, unwrap=False)
        if start_wrapper is None:
            continue
        stop_wrapper = _inspect._get_effective(leaf, StopBeam, unwrap=False)
        if stop_wrapper is None:
            violators.append(leaf)
            continue
        if stop_wrapper.leaked_start_offset < start_wrapper.leaked_start_offset:
            violators.append(leaf)
            continue
        leaf_start_offset = leaf._get_timespan().start_offset
        if stop_wrapper.leaked_start_offset == leaf_start_offset:
            violators.append(leaf)
    return violators, total


def check_duplicate_ids(argument) -> typing.Tuple[typing.List, int]:
    """
    Checks duplicate IDs.
    """
    violators = []
    components = Iteration(argument).components()
    total_ids = [id(_) for _ in components]
    unique_ids = Sequence(total_ids).remove_repeats()
    if len(unique_ids) < len(total_ids):
        for current_id in unique_ids:
            if 1 < total_ids.count(current_id):
                violators.extend([_ for _ in components if id(_) == current_id])
    return violators, len(total_ids)


def check_empty_containers(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks empty containers.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> staff.append(abjad.Container())

        >>> print(abjad.lilypond(staff))
        \new Staff
        {
            c'4
            d'4
            e'4
            f'4
            {
            }
        }

        >>> violators, total = abjad.wf.check_empty_containers(staff)
        >>> violators
        [Container()]

    """
    violators, containers = [], set()
    for container in Iteration(argument).components(Container):
        containers.add(container)
        if len(container) == 0:
            violators.append(container)
    return violators, len(containers)


def check_missing_parents(argument) -> typing.Tuple[typing.List, int]:
    """
    Checks missing parents.
    """
    violators, total = [], set()
    components = Iteration(argument).components()
    for i, component in enumerate(components):
        total.add(component)
        if 0 < i:
            # TODO: figure out why "if component._parent is None" doesn't work
            if Parentage(component).parent is None:
                violators.append(component)
    return violators, len(total)


def check_notes_on_wrong_clef(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks notes and chords on wrong clef.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> clef = abjad.Clef('alto')
        >>> abjad.attach(clef, staff[0])
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(staff))
            \new Staff
            {
                \clef "alto"
                c'8
                d'8
                e'8
                f'8
            }

        >>> print(abjad.wf.tabulate_wellformedness(staff))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        4 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

    ..  container:: example

        Always allows percussion clef:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> clef = abjad.Clef('percussion')
        >>> abjad.attach(clef, staff[0])
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(staff))
            \new Staff
            {
                \clef "percussion"
                c'8
                d'8
                e'8
                f'8
            }

        >>> print(abjad.wf.tabulate_wellformedness(staff))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

    """
    violators, total = [], set()
    for leaf in Iteration(argument).leaves():
        total.add(leaf)
        instrument = _inspect._get_effective(leaf, Instrument)
        if instrument is None:
            continue
        clef = _inspect._get_effective(leaf, Clef)
        if clef is None:
            continue
        allowable_clefs = [Clef(_) for _ in instrument.allowable_clefs]
        allowable_clefs.append(Clef("percussion"))
        if clef not in allowable_clefs:
            violators.append(leaf)
    return violators, len(total)


def check_out_of_range_pitches(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks out-of-range notes.

    ..  container:: example

        Out of range:

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(staff))
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        >>> print(abjad.wf.tabulate_wellformedness(staff))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        1 /     2 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

    ..  container:: example

        Allows out-of-range pitches:

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.attach("ALLOW_OUT_OF_RANGE", staff[2])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(staff))
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        >>> print(abjad.wf.tabulate_wellformedness(staff))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     2 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

    """
    violators, total = [], set()
    for leaf in Iteration(argument).leaves(pitched=True):
        total.add(leaf)
        if leaf._has_indicator("ALLOW_OUT_OF_RANGE"):
            continue
        if leaf._has_indicator("HIDDEN"):
            continue
        if "unpitched" in argument._get_indicators(str):
            continue
        instrument = _inspect._get_effective(leaf, Instrument)
        if instrument is None:
            continue
        # if leaf not in instrument.pitch_range:
        if not sounding_pitches_are_in_range(leaf, instrument.pitch_range):
            violators.append(leaf)
    return violators, len(total)


def check_overlapping_text_spanners(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks overlapping text spanners.

    ..  container:: example

        Overlapping text spanners are not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> abjad.text_spanner(voice)
        >>> abjad.text_spanner(voice[1:3])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            \startTextSpan
            c'4
            \stopTextSpan
            c'4
            \stopTextSpan
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        1 /     2 overlapping text spanners
        0 /     2 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     2 unterminated text spanners

    ..  container:: example

        Overlapping text spanners with different IDs are wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> abjad.text_spanner(voice)
        >>> command = r'\startTextSpanOne'
        >>> start_text_span = abjad.StartTextSpan(command=command)
        >>> abjad.attach(start_text_span, voice[1])
        >>> command = r'\stopTextSpanOne'
        >>> stop_text_span = abjad.StopTextSpan(command=command)
        >>> abjad.attach(stop_text_span, voice[2])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            \startTextSpanOne
            c'4
            \stopTextSpanOne
            c'4
            \stopTextSpan
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     2 overlapping text spanners
        0 /     2 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     2 unterminated text spanners

    ..  container:: example

        Enchained text spanners do not overlap (and are wellformed):

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> abjad.text_spanner(voice[:3])
        >>> abjad.text_spanner(voice[-2:])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            \stopTextSpan
            \startTextSpan
            c'4
            \stopTextSpan
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     2 overlapping text spanners
        0 /     2 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     2 unterminated text spanners

    ..  container:: example

        REGRESSION. Matching start- and stop-text-spans on a single leaf do
        not overlap (and are wellformed) iff stop-text-span leaks to the
        right:

        >>> voice = abjad.Voice("c'2 d'2 e'2 f'2")
        >>> abjad.attach(abjad.StartTextSpan(), voice[0])
        >>> stop_text_span = abjad.StopTextSpan(leak=True)
        >>> abjad.attach(stop_text_span, voice[0])
        >>> abjad.attach(abjad.StartTextSpan(), voice[2])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[3])
        >>> abjad.show(voice) # doctest: +SKIP

        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'2
            \startTextSpan
            <> \stopTextSpan
            d'2
            e'2
            \startTextSpan
            f'2
            \stopTextSpan
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     2 overlapping text spanners
        0 /     2 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     2 unterminated text spanners

    """
    violators, total = [], 0

    def key(wrapper):
        if isinstance(wrapper.indicator, StartTextSpan):
            priority = 1
        else:
            priority = 0
        return (wrapper.leaked_start_offset, priority)

    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        wrappers.sort(key=key)
        open_spanners: typing.Dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.indicator, StartTextSpan):
                total += 1
                command = wrapper.indicator.command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                if open_spanners[command]:
                    violators.append(wrapper.component)
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.indicator, StopTextSpan):
                command = wrapper.indicator.command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command in open_spanners and open_spanners[command]:
                    open_spanners[command].pop()
    return violators, total


def check_unmatched_stop_text_spans(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks unmatched stop text spans.

    ..  container:: example

        Unmatched stop text span is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            c'4
            c'4
            c'4
            \stopTextSpan
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        1 /     0 unmatched stop text spans
        0 /     0 unterminated hairpins
        0 /     0 unterminated text spanners

        Matched stop text span is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'4
                \startTextSpan
                c'4
                c'4
                c'4
                \stopTextSpan
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        open_spanners: typing.Dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.indicator, StartTextSpan):
                total += 1
                command = wrapper.indicator.command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.indicator, StopTextSpan):
                command = wrapper.indicator.command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command not in open_spanners or not open_spanners[command]:
                    violators.append(wrapper.component)
                else:
                    open_spanners[command].pop()
    return violators, total


def check_unterminated_hairpins(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks unterminated hairpins.

    ..  container:: example

        Unterminated crescendo is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            \<
            c'4
            c'4
            c'4
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        1 /     1 unterminated hairpins
        0 /     0 unterminated text spanners

        Even with start dynamic:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> dynamic = abjad.Dynamic('f')
        >>> abjad.attach(dynamic, voice[0])
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            \f
            \<
            c'4
            c'4
            c'4
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     0 overlapping text spanners
        0 /     0 unmatched stop text spans
        1 /     1 unterminated hairpins
        0 /     0 unterminated text spanners

        Terminated crescendo is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> dynamic = abjad.Dynamic('f')
        >>> abjad.attach(dynamic, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'4
                \<
                c'4
                c'4
                c'4
                \f
            }

        >>> abjad.wf.wellformed(voice)
        True

    ..  container:: example

        Bang-terminated crescendo is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> stop_hairpin = abjad.StopHairpin()
        >>> abjad.attach(stop_hairpin, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'4
                \<
                c'4
                c'4
                c'4
                \!
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        last_dynamic = None
        last_tag = None
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        for wrapper in wrappers:
            parameter = getattr(wrapper.indicator, "parameter", None)
            if parameter == "DYNAMIC" or isinstance(wrapper.indicator, StopHairpin):
                last_dynamic = wrapper.indicator
                last_tag = wrapper.tag
                if isinstance(wrapper.indicator, StartHairpin):
                    total += 1
        if isinstance(last_dynamic, StartHairpin) and str(
            Tag("RIGHT_BROKEN")
        ) not in str(last_tag):
            violators.append(wrapper.component)
    return violators, total


def check_unterminated_text_spanners(argument) -> typing.Tuple[typing.List, int]:
    r"""
    Checks unterminated text spanners.

    ..  container:: example

        Unterminated text spanner is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> print(abjad.lilypond(voice))
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            c'4
        }

        >>> print(abjad.wf.tabulate_wellformedness(voice))
        0 /     4 beamed long notes
        0 /     5 duplicate ids
        0 /     1 empty containers
        0 /     5 missing parents
        0 /     4 notes on wrong clef
        0 /     4 out of range pitches
        0 /     1 overlapping text spanners
        0 /     1 unmatched stop text spans
        0 /     0 unterminated hairpins
        1 /     1 unterminated text spanners

        Terminated text span is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> print(abjad.lilypond(voice))
            \new Voice
            {
                c'4
                \startTextSpan
                c'4
                c'4
                c'4
                \stopTextSpan
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        open_spanners: typing.Dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.indicator, StartTextSpan):
                total += 1
                command = wrapper.indicator.command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.indicator, StopTextSpan):
                command = wrapper.indicator.command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command not in open_spanners or not open_spanners[command]:
                    # unmatched stop text span
                    pass
                else:
                    open_spanners[command].pop()
        for command, list_ in open_spanners.items():
            for component in list_:
                violators.append(component)
    return violators, total


_globals = globals()


def _call_functions(
    component,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
):
    triples = []
    if check_beamed_long_notes:
        name = "check_beamed_long_notes"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_duplicate_ids:
        name = "check_duplicate_ids"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_empty_containers:
        name = "check_empty_containers"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_missing_parents:
        name = "check_missing_parents"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_notes_on_wrong_clef:
        name = "check_notes_on_wrong_clef"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_out_of_range_pitches:
        name = "check_out_of_range_pitches"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_overlapping_text_spanners:
        name = "check_overlapping_text_spanners"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unmatched_stop_text_spans:
        name = "check_unmatched_stop_text_spans"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unterminated_hairpins:
        name = "check_unterminated_hairpins"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unterminated_text_spanners:
        name = "check_unterminated_text_spanners"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    return triples


def tabulate_wellformedness(
    component,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
) -> str:
    """
    Tabulates wellformedness.
    """
    triples = _call_functions(
        component,
        check_beamed_long_notes=check_beamed_long_notes,
        check_duplicate_ids=check_duplicate_ids,
        check_empty_containers=check_empty_containers,
        check_missing_parents=check_missing_parents,
        check_notes_on_wrong_clef=check_notes_on_wrong_clef,
        check_out_of_range_pitches=check_out_of_range_pitches,
        check_overlapping_text_spanners=check_overlapping_text_spanners,
        check_unmatched_stop_text_spans=check_unmatched_stop_text_spans,
        check_unterminated_hairpins=check_unterminated_hairpins,
        check_unterminated_text_spanners=check_unterminated_text_spanners,
    )
    strings = []
    for violators, total, name in triples:
        violator_count = len(violators)
        name = name.replace("check_", "")
        name = name.replace("_", " ")
        string = f"{violator_count} /    {total} {name}"
        strings.append(string)
    return "\n".join(strings)


def wellformed(
    component,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
) -> bool:
    """
    Is true when ``component`` is wellformed.
    """
    triples = _call_functions(
        component,
        check_beamed_long_notes=check_beamed_long_notes,
        check_duplicate_ids=check_duplicate_ids,
        check_empty_containers=check_empty_containers,
        check_missing_parents=check_missing_parents,
        check_notes_on_wrong_clef=check_notes_on_wrong_clef,
        check_out_of_range_pitches=check_out_of_range_pitches,
        check_overlapping_text_spanners=check_overlapping_text_spanners,
        check_unmatched_stop_text_spans=check_unmatched_stop_text_spans,
        check_unterminated_hairpins=check_unterminated_hairpins,
        check_unterminated_text_spanners=check_unterminated_text_spanners,
    )
    for violators, total, name in triples:
        if violators:
            return False
    return True



import abc
import collections.abc

from .new import new
from .storage import FormatSpecification, StorageFormatManager


class TypedCollection:
    """
    Abstract typed collection.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_collection", "_item_class")

    _is_abstract = True

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, items=None, item_class=None):
        assert isinstance(item_class, (type(None), type))
        self._item_class = item_class

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        """
        Is true when typed collection contains ``item``.

        Returns true or false.
        """
        try:
            item = self._coerce_item(item)
        except ValueError:
            return False
        return self._collection.__contains__(item)

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a typed collection with items that
        compare equal to those of this typed collection.

        Returns true or false.
        """
        if issubclass(type(argument), type(self)):
            return self._collection == argument._collection
        elif isinstance(argument, type(self._collection)):
            return self._collection == argument
        return False

    def __hash__(self):
        """
        Hashes typed collection.

        Redefined in tandem with __eq__.
        """
        return object.__hash__(self)

    def __iter__(self):
        """
        Iterates typed collection.

        Returns generator.
        """
        return self._collection.__iter__()

    def __len__(self):
        """
        Gets length of typed collection.

        Returns nonnegative integer.
        """
        return len(self._collection)

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _coerce_item(self, item):
        def coerce_(x):
            if isinstance(x, self._item_class):
                return x
            return self._item_class(x)

        if self._item_class is None:
            return item
        return coerce_(item)

    def _get_format_specification(self):
        names = list(StorageFormatManager(self).signature_keyword_names)
        if "items" in names:
            names.remove("items")
        return FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=[self._collection],
            storage_format_keyword_names=names,
        )

    def _on_insertion(self, item):
        """
        Override to operate on item after insertion into collection.
        """
        pass

    def _on_removal(self, item):
        """
        Override to operate on item after removal from collection.
        """
        pass

    ### PUBLIC PROPERTIES ###

    @property
    def item_class(self):
        """
        Gets item class of collection.

        Collection coerces items according to ``item_class``.

        Returns class.
        """
        return self._item_class

    @property
    def items(self):
        """
        Gets items in collection.

        Returns list.
        """
        return [_ for _ in self]


class TypedCounter(TypedCollection, collections.abc.MutableMapping):
    """
    Typed counter.

    ..  container:: example

        >>> counter = abjad.TypedCounter(
        ...     [0, "c'", 1, True, "cs'", "df'"],
        ...     item_class=abjad.NumberedPitch,
        ...     )

        >>> string = abjad.storage(counter)
        >>> print(string)
        abjad.TypedCounter(
            {
                abjad.NumberedPitch(0): 2,
                abjad.NumberedPitch(1): 4,
                },
            item_class=abjad.NumberedPitch,
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None, **keywords):
        TypedCollection.__init__(self, item_class=item_class, items=items)
        self._collection = collections.Counter()
        self.update(items, **keywords)

    ### SPECIAL METHODS ###

    def __add__(self, argument):
        """
        Adds typed counter to ``argument``.

        Returns new typed counter.
        """
        if (
            not isinstance(argument, type(self))
            or not self.item_class == argument.item_class
        ):
            return NotImplemented
        result = type(self)()
        result._collection = self._collection + argument._collection
        return result

    def __and__(self, argument):
        """
        Logical AND of typed counter and ``argument``.

        Returns new typed counter.
        """
        if (
            not isinstance(argument, type(self))
            or not self.item_class == argument.item_class
        ):
            return NotImplemented
        result = type(self)()
        result._collection = self._collection & argument._collection
        return result

    def __delitem__(self, item):
        """
        Deletes ``item`` from typed counter.

        Returns none.
        """
        item = self._coerce_item(item)
        if item in self._collection:
            dict.__delitem__(self._collection, item)

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        Returns item or slice.
        """
        argument = self._coerce_item(argument)
        return self._collection.__getitem__(argument)

    def __or__(self, argument):
        """
        Logical OR of typed counter and ``argument``.

        Returns new typed counter.
        """
        if (
            not isinstance(argument, type(self))
            or not self.item_class == argument.item_class
        ):
            return NotImplemented
        result = type(self)()
        result._collection = self._collection | argument._collection
        return result

    def __radd__(self, argument):
        """
        Adds ``argument`` to typed counter.

        Returns new typed counter.
        """
        if (
            not isinstance(argument, type(self))
            or not self.item_class == argument.item_class
        ):
            return NotImplemented
        result = type(self)()
        result._collection = argument._collection + self._collection
        return result

    def __reduce__(self):
        """
        Reduces typed counter.

        Returns new typed counter.
        """
        return type(self), (dict(self._collection),)

    def __setitem__(self, item, value):
        """
        Sets typed counter ``item`` to ``value``.

        Returns none.
        """
        item = self._coerce_item(item)
        self._collection.__setitem__(item, value)

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from typed counter.

        Returns new typed counter.
        """
        if (
            not isinstance(argument, type(self))
            or not self.item_class == argument.item_class
        ):
            return NotImplemented
        result = type(self)()
        result._collection = self._collection - argument._collection
        return result

    ### PRIVATE METHODS ###

    def _coerce_arguments(self, items=None, **keywords):
        def _coerce_mapping(items):
            the_items = {}
            for item, count in items.items():
                item = self._coerce_item(item)
                if item not in the_items:
                    the_items[item] = 0
                the_items[item] += count
            return the_items

        the_items = []
        if items is not None:
            if isinstance(items, collections.abc.Mapping):
                items = _coerce_mapping(items)
            else:
                the_items = []
                for item in items:
                    the_items.append(self._coerce_item(item))
        itemdict = _coerce_mapping(keywords)
        return the_items, itemdict

    def _get_format_specification(self):
        names = list(StorageFormatManager(self).signature_keyword_names)
        if "items" in names:
            names.remove("items")
        return FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=[self._collection],
            storage_format_keyword_names=names,
            template_names=names,
        )

    ### PUBLIC METHODS ###

    def clear(self):
        """
        Clears typed counter.

        Returns none.
        """
        self._collection.clear()

    def copy(self):
        """
        Copies typed counter.

        Returns new typed counter.
        """
        return type(self)(self)

    def elements(self):
        """
        Elements in typed counter.
        """
        return self._collection.elements()

    @classmethod
    def fromkeys(class_, iterable, v=None):
        """
        Makes new typed counter from ``iterable``.

        Not yet impelemented.

        Will return new typed counter.
        """
        name = class_.__name__
        message = f"{name}.fromkeys() is undefined. Use {name}(iterable) instead."
        raise NotImplementedError(message)

    def items(self):
        """
        Items in typed counter.

        Returns tuple.
        """
        return list(self._collection.items())

    def keys(self):
        """
        Iterates keys in typed counter.
        """
        return iter(self._collection.keys())

    def most_common(self, n=None):
        """
        Please document.
        """
        return self._collection(n=n)

    def subtract(self, iterable=None, **keywords):
        """
        Subtracts ``iterable`` from typed counter.
        """
        items, itemdict = self._coerce_arguments(iterable, **keywords)
        self._collection.subtract(items, **itemdict)

    def update(self, iterable=None, **keywords):
        """
        Updates typed counter with ``iterable``.
        """
        items, itemdict = self._coerce_arguments(iterable, **keywords)
        self._collection.update(items, **itemdict)

    def values(self):
        """
        Iterates values in typed counter.
        """
        return iter(self._collection.values())

    def viewitems(self):
        """
        Please document.
        """
        return self._collection.items()

    def viewkeys(self):
        """
        Please document.
        """
        return self._collection.keys()

    def viewvalues(self):
        """
        Please document.
        """
        return self._collection.values()


class TypedFrozenset(TypedCollection, collections.abc.Set):
    """
    Typed fozen set.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        TypedCollection.__init__(self, item_class=item_class, items=items)
        items = items or []
        items = [self._coerce_item(_) for _ in items]
        self._collection = frozenset(items)

    ### SPECIAL METHODS ###

    def __and__(self, argument):
        """
        Logical AND of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__and__(argument._collection)
        result = type(self)(result)
        return result

    def __ge__(self, argument):
        """
        Is true when typed frozen set is greater than or equal to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__ge__(argument._collection)

    def __gt__(self, argument):
        """
        Is true when typed frozen set is greater than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__gt__(argument._collection)

    def __hash__(self):
        """
        Hashes typed frozen set.

        Returns integer.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __le__(self, argument):
        """
        Is true when typed frozen set is less than or equal to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__le__(argument._collection)

    def __lt__(self, argument):
        """
        Is true when typed frozen set is less than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__lt__(argument._collection)

    def __or__(self, argument):
        """
        Logical OR of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__or__(argument._collection)
        result = type(self)(result)
        return result

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from typed frozen set.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__sub__(argument._collection)
        result = type(self)(result)
        return result

    def __xor__(self, argument):
        """
        Logical XOR of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__xor__(argument._collection)
        result = type(self)(result)
        return result

    ### PUBLIC METHODS ###

    def copy(self):
        """
        Copies typed frozen set.

        Returns new typed frozen set.
        """
        return type(self)(self._collection.copy())

    def difference(self, argument):
        """
        Typed frozen set set-minus ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.difference(argument._collection)
        result = type(self)(result)
        return result

    def intersection(self, argument):
        """
        Set-theoretic intersection of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.intersection(argument._collection)
        result = type(self)(result)
        return result

    def isdisjoint(self, argument):
        """
        Is true when typed frozen set shares no elements with ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.isdisjoint(argument._collection)

    def issubset(self, argument):
        """
        Is true when typed frozen set is a subset of ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.issubset(argument._collection)

    def issuperset(self, argument):
        """
        Is true when typed frozen set is a superset of ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.issuperset(argument._collection)

    def symmetric_difference(self, argument):
        """
        Symmetric difference of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.symmetric_difference(argument._collection)
        result = type(self)(result)
        return result

    def union(self, argument):
        """
        Union of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.union(argument._collection)
        result = type(self)(result)
        return result


class TypedList(TypedCollection, collections.abc.MutableSequence):
    """
    Typed list.

    ..  container:: example

        No item coercion:

        >>> list_ = abjad.TypedList()
        >>> list_.append(23)
        >>> list_.append('foo')
        >>> list_.append(False)
        >>> list_.append((1, 2, 3))
        >>> list_.append(3.14159)

        >>> string = abjad.storage(list_)
        >>> print(string)
        abjad.TypedList(
            [
                23,
                'foo',
                False,
                (1, 2, 3),
                3.14159,
                ]
            )

    ..  container:: example

        Named pitch item coercion:

        >>> pitch_list = abjad.TypedList(
        ...     item_class=abjad.NamedPitch,
        ...     )
        >>> pitch_list.append(0)
        >>> pitch_list.append("d'")
        >>> pitch_list.append(('e', 4))
        >>> pitch_list.append(abjad.NamedPitch("f'"))

        >>> string = abjad.storage(pitch_list)
        >>> print(string)
        abjad.TypedList(
            [
                abjad.NamedPitch("c'"),
                abjad.NamedPitch("d'"),
                abjad.NamedPitch("e'"),
                abjad.NamedPitch("f'"),
                ],
            item_class=abjad.NamedPitch,
            )

    Ordered collection with optional item coercion.

    Implements the list interface.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_keep_sorted",)

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None, keep_sorted=False):
        TypedCollection.__init__(self, item_class=item_class, items=items)
        self._collection = []
        if keep_sorted is not None:
            assert isinstance(keep_sorted, bool), repr(keep_sorted)
        self._keep_sorted = keep_sorted
        items = items or []
        the_items = []
        for item in items:
            the_items.append(self._coerce_item(item))
        self.extend(the_items)

    ### SPECIAL METHODS ###

    def __delitem__(self, i):
        """
        Aliases list.__delitem__().

        Returns none.
        """
        self._on_removal(self._collection[i])
        del self._collection[i]

    def __getitem__(self, argument):
        """
        Gets item or slice identified  by ``argument``.

        Returns item or slice.
        """
        return self._collection.__getitem__(argument)

    def __iadd__(self, argument):
        """
        Adds ``argument`` in place to typed list.

        ..  container:: example

            >>> dynamic_list = abjad.TypedList(item_class=abjad.Dynamic)
            >>> dynamic_list.append('ppp')
            >>> dynamic_list += ['p', 'mp', 'mf', 'fff']

            >>> string = abjad.storage(dynamic_list)
            >>> print(string)
            abjad.TypedList(
                [
                    abjad.Dynamic('ppp'),
                    abjad.Dynamic('p'),
                    abjad.Dynamic('mp'),
                    abjad.Dynamic('mf'),
                    abjad.Dynamic('fff'),
                    ],
                item_class=abjad.Dynamic,
                )

        Returns typed list.
        """
        self.extend(argument)
        return self

    def __reversed__(self):
        """
        Aliases list.__reversed__().

        Returns generator.
        """
        return self._collection.__reversed__()

    def __setitem__(self, i, argument):
        """
        Sets item ``i`` equal to ``argument``.

        ..  container:: example

            Sets item:

            >>> pitch_list = abjad.TypedList(
            ...     item_class=abjad.NamedPitch,
            ...     )
            >>> pitch_list.append(0)
            >>> pitch_list.append("d'")
            >>> pitch_list.append(('e', 4))
            >>> pitch_list.append(abjad.NamedPitch("f'"))

            >>> pitch_list[-1] = 'gqs,'
            >>> string = abjad.storage(pitch_list)
            >>> print(string)
            abjad.TypedList(
                [
                    abjad.NamedPitch("c'"),
                    abjad.NamedPitch("d'"),
                    abjad.NamedPitch("e'"),
                    abjad.NamedPitch('gqs,'),
                    ],
                item_class=abjad.NamedPitch,
                )

        ..  container:: example

            Sets slice:

            >>> pitch_list[-1:] = ["f'", "g'", "a'", "b'", "c''"]
            >>> string = abjad.storage(pitch_list)
            >>> print(string)
            abjad.TypedList(
                [
                    abjad.NamedPitch("c'"),
                    abjad.NamedPitch("d'"),
                    abjad.NamedPitch("e'"),
                    abjad.NamedPitch("f'"),
                    abjad.NamedPitch("g'"),
                    abjad.NamedPitch("a'"),
                    abjad.NamedPitch("b'"),
                    abjad.NamedPitch("c''"),
                    ],
                item_class=abjad.NamedPitch,
                )

        Returns none.
        """
        if isinstance(i, int):
            new_item = self._coerce_item(argument)
            old_item = self._collection[i]
            self._on_removal(old_item)
            self._on_insertion(new_item)
            self._collection[i] = new_item
        elif isinstance(i, slice):
            new_items = [self._coerce_item(item) for item in argument]
            old_items = self._collection[i]
            for old_item in old_items:
                self._on_removal(old_item)
            for new_item in new_items:
                self._on_insertion(new_item)
            self._collection[i] = new_items
        if self.keep_sorted:
            self.sort()

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        names = list(StorageFormatManager(self).signature_keyword_names)
        if "items" in names:
            names.remove("items")
        if "keep_sorted" in names:
            names.remove("keep_sorted")
        return FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=[self._collection],
            storage_format_keyword_names=names,
        )

    ### PUBLIC METHODS ###

    def append(self, item):
        """
        Appends ``item`` to typed list.

        ..  container:: example

            >>> integer_list = abjad.TypedList(item_class=int)
            >>> integer_list.append('1')
            >>> integer_list.append(2)
            >>> integer_list.append(3.4)
            >>> integer_list[:]
            [1, 2, 3]

        Returns none.
        """
        item = self._coerce_item(item)
        self._on_insertion(item)
        self._collection.append(item)
        if self.keep_sorted:
            self.sort()

    def count(self, item):
        """
        Gets count of ``item`` in typed list.

        ..  container:: example

            >>> integer_list = abjad.TypedList(item_class=int)
            >>> integer_list.extend([0, 0., '0', 99])

            >>> integer_list.count(0)
            3
            >>> integer_list.count(1)
            0
            >>> integer_list.count(99)
            1

        Returns nonnegative integer.
        """
        item = self._coerce_item(item)
        return self._collection.count(item)

    def extend(self, items):
        """
        Extends typed list with ``items``.

        ..  container:: example

            >>> integer_list = abjad.TypedList(item_class=int)
            >>> integer_list.extend(['0', 1.0, 2, 3.14159])
            >>> integer_list
            TypedList([0, 1, 2, 3], item_class=int)

        Returns none.
        """
        for item in items:
            self.append(item)
        if self.keep_sorted:
            self.sort()

    def index(self, item):
        """
        Gets index of ``item`` in typed list.

        ..  container:: example

            >>> pitch_list = abjad.TypedList(
            ...     item_class=abjad.NamedPitch,
            ...     )
            >>> pitch_list.extend(['cqf', "as'", 'b,', 'dss'])

            >>> pitch_list.index(abjad.NamedPitch('cqf'))
            0
            >>> pitch_list.index(abjad.NamedPitch("as'"))
            1
            >>> pitch_list.index('b,')
            2
            >>> pitch_list.index('dss')
            3

        Returns nonnegative integer.
        """
        item = self._coerce_item(item)
        return self._collection.index(item)

    def insert(self, i, item):
        """
        Insert ``item`` into typed list.

        ..  container:: example

            Inserts into typed list.

            >>> integer_list = abjad.TypedList(item_class=int)
            >>> integer_list.extend(['1', 2, 4.3])
            >>> integer_list
            TypedList([1, 2, 4], item_class=int)

            >>> integer_list.insert(0, '0')
            >>> integer_list
            TypedList([0, 1, 2, 4], item_class=int)

            >>> integer_list.insert(1, '9')
            >>> integer_list
            TypedList([0, 9, 1, 2, 4], item_class=int)

        Returns none.
        """
        item = self._coerce_item(item)
        self._on_insertion(item)
        result = self._collection.insert(i, item)
        if self.keep_sorted:
            self.sort()
        return result

    def pop(self, i=-1):
        """
        Pops item ``i`` from typed list.

        Returns item.
        """
        result = self._collection.pop(i)
        self._on_removal(result)
        if self.keep_sorted:
            self.sort()
        return result

    def remove(self, item):
        """
        Removes ``item`` from typed list.

        ..  container:: example

            >>> integer_list = abjad.TypedList(item_class=int)
            >>> integer_list.extend(('0', 1.0, 2, 3.14159))
            >>> integer_list[:]
            [0, 1, 2, 3]

            >>> integer_list.remove('1')
            >>> integer_list[:]
            [0, 2, 3]

        Returns none.
        """
        item = self._coerce_item(item)
        index = self._collection.index(item)
        item = self._collection[index]
        self._on_removal(item)
        del self._collection[index]
        if self.keep_sorted:
            self.sort()

    def reverse(self):
        """
        Reverses items in typed list.
        """
        self._collection.reverse()

    def sort(self, cmp=None, key=None, reverse=False):
        """
        Sorts items in typed list.
        """
        if cmp is not None:

            def cmp_to_key(comparator):
                """
                Convert a ``cmp`` function into a ``key`` function for use
                with ``sort()``.
                """

                class CmpToKey:
                    def __init__(self, argument):
                        self.argument = argument

                    def __lt__(self, other):
                        return comparator(self.argument, other.argument) < 0

                    def __gt__(self, other):
                        return comparator(self.argument, other.argument) > 0

                    def __eq__(self, other):
                        return comparator(self.argument, other.argument) == 0

                    def __le__(self, other):
                        return comparator(self.argument, other.argument) <= 0

                    def __ge__(self, other):
                        return comparator(self.argument, other.argument) >= 0

                    def __ne__(self, other):
                        return comparator(self.argument, other.argument) != 0

                return CmpToKey

            key = cmp_to_key(cmp)
        self._collection.sort(key=key, reverse=reverse)

    ### PUBLIC PROPERTIES ###

    @property
    def keep_sorted(self):
        """
        Is true when typed list keeps items sorted.

        Defaults to none.

        Set to true, false or none.

        Returns true, false or none.
        """
        return self._keep_sorted

    @keep_sorted.setter
    def keep_sorted(self, argument):
        assert isinstance(argument, (bool, type(None)))
        self._keep_sorted = argument


class TypedTuple(TypedCollection, collections.abc.Sequence):
    """
    Typed tuple.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        TypedCollection.__init__(self, item_class=item_class, items=items)
        items = items or []
        self._collection = tuple(self._coerce_item(item) for item in items)

    ### SPECIAL METHODS ###

    def __add__(self, argument):
        """
        Adds typed tuple to ``argument``.

        Returns new typed tuple.
        """

        if isinstance(argument, type(self)):
            items = argument._collection
            return new(self, items=self._collection[:] + items)
        elif isinstance(argument, type(self._collection)):
            items = argument[:]
            return new(self, items=self._collection[:] + items)
        raise NotImplementedError

    def __contains__(self, item):
        """
        Is true if typed tuple contains ``item``.

        Coerces ``item``.

        Returns none.
        """
        try:
            item = self._coerce_item(item)
        except ValueError:
            return False
        return self._collection.__contains__(item)

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        Returns item or new typed tuple.
        """
        item = self._collection.__getitem__(argument)
        try:
            return type(self)(item)
        except TypeError:
            return item

    def __hash__(self):
        """
        Hashes typed tuple.

        Returns integer.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __mul__(self, argument):
        """
        Multiplies typed tuple by ``argument``.

        Returns new typed tuple.
        """
        items = self._collection * argument
        return new(self, items=items)

    def __radd__(self, argument):
        """
        Right-adds ``argument`` to typed tuple.
        """
        items = argument + self._collection
        return new(self, items=items)

    def __rmul__(self, argument):
        """
        Multiplies ``argument`` by typed tuple.

        Returns new typed tuple.
        """
        return self.__mul__(argument)

    ### PUBLIC METHODS ###

    def count(self, item):
        """
        Counts ``item`` in collection.

        Coerces ``item``.

        Returns nonnegative integer.
        """
        try:
            item = self._coerce_item(item)
        except (ValueError, TypeError):
            return 0
        return self._collection.count(item)

    def index(self, item):
        """
        Gets index of ``item`` in collection.

        Coerces ``item``.

        Returns nonnegative integer.
        """
        item = self._coerce_item(item)
        return self._collection.index(item)



import abc
import collections
import copy
import types

from .. import enumerate
from ..new import new
from ..sequence import Sequence
from ..storage import FormatSpecification
from ..typedcollections import TypedCollection, TypedFrozenset
from .intervalclasses import IntervalClass, NamedIntervalClass, NumberedIntervalClass
from .intervals import Interval, NamedInterval, NumberedInterval
from .pitchclasses import NamedPitchClass, NumberedPitchClass, PitchClass
from .pitches import NamedPitch, NumberedPitch, Pitch
from .segments import PitchClassSegment, PitchSegment


class Set(TypedFrozenset):
    """
    Abstract set.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_expression",)

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        if isinstance(items, str):
            items = items.split()
        elif isinstance(items, (collections.abc.Iterator, types.GeneratorType)):
            items = [item for item in items]
        if item_class is None:
            item_class = self._named_item_class
            if items is not None:
                if isinstance(items, TypedCollection) and issubclass(
                    items.item_class, self._parent_item_class
                ):
                    item_class = items.item_class
                elif len(items):
                    if isinstance(items, collections.abc.Set):
                        items = tuple(items)
                    if isinstance(items[0], str):
                        item_class = self._named_item_class
                    elif isinstance(items[0], (int, float)):
                        item_class = self._numbered_item_class
                    elif isinstance(items[0], self._parent_item_class):
                        item_class = type(items[0])
        assert issubclass(item_class, self._parent_item_class)
        TypedFrozenset.__init__(self, items=items, item_class=item_class)
        self._expression = None

    ### SPECIAL METHODS ###

    def __str__(self):
        """
        Gets string representation.

        Returns string.
        """
        items = self._get_sorted_repr_items()
        items = [str(_) for _ in items]
        string = ", ".join(items)
        return f"{{{string}}}"

    ### PRIVATE PROPERTIES ###

    @abc.abstractproperty
    def _named_item_class(self):
        raise NotImplementedError

    @abc.abstractproperty
    def _numbered_item_class(self):
        raise NotImplementedError

    @abc.abstractproperty
    def _parent_item_class(self):
        raise NotImplementedError

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        repr_items = self._get_sorted_repr_items()
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            repr_args_values=[repr_items],
            storage_format_args_values=[repr_items],
            storage_format_keyword_names=[],
        )

    def _get_sorted_repr_items(self):
        items = sorted(self, key=lambda x: (float(x.number), str(x)))
        if self.item_class.__name__.startswith("Named"):
            repr_items = [str(x) for x in items]
        elif hasattr(self.item_class, "number"):
            repr_items = [x.number for x in items]
        elif hasattr(self.item_class, "pitch_class_number"):
            repr_items = [x.pitch_class_number for x in items]
        elif hasattr(self.item_class, "__abs__"):
            repr_items = [abs(x) for x in items]
        else:
            raise ValueError(f"invalid item class: {self.item_class!r}.")
        return repr_items

    def _sort_self(self):
        return tuple(self)

    ### PUBLIC PROPERTIES ###

    @property
    def cardinality(self):
        """
        Gets cardinality of set.

        Defined equal to length of set.

        Returns nonnegative integer.
        """
        return len(self)

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes set from ``selection``.

        Returns set.
        """
        raise NotImplementedError


class IntervalClassSet(Set):
    """
    Interval-class set.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        prototype = (
            PitchClassSegment,
            PitchSegment,
            PitchClassSet,
            PitchSet,
        )
        if isinstance(items, prototype):
            items = list(items)
            pairs = enumerate.yield_pairs(items)
            items = [second - first for first, second in pairs]
        super().__init__(items=items, item_class=item_class)

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedIntervalClass

    @property
    def _numbered_item_class(self):
        return NumberedIntervalClass

    @property
    def _parent_item_class(self):
        return IntervalClass

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        r"""
        Initialize interval set from component selection:

        ..  container:: example

            ::

                >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
                >>> staff_2 = abjad.Staff("c4. r8 g2")
                >>> staff_group = abjad.StaffGroup([staff_1, staff_2])
                >>> abjad.show(staff_group) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff_group)
                >>> print(string)
                \new StaffGroup
                <<
                    \new Staff
                    {
                        c'4
                        <d' fs' a'>4
                        b2
                    }
                    \new Staff
                    {
                        c4.
                        r8
                        g2
                    }
                >>

            ::

                >>> selection = abjad.select(staff_group)
                >>> interval_classes = abjad.IntervalClassSet.from_selection(selection)
                >>> for interval_class in sorted(interval_classes):
                ...     interval_class
                ...
                NamedIntervalClass('-M6')
                NamedIntervalClass('-P5')
                NamedIntervalClass('-A4')
                NamedIntervalClass('-M3')
                NamedIntervalClass('-m3')
                NamedIntervalClass('-M2')
                NamedIntervalClass('+m2')
                NamedIntervalClass('+M2')
                NamedIntervalClass('+m3')
                NamedIntervalClass('+M3')
                NamedIntervalClass('+P4')
                NamedIntervalClass('+A4')
                NamedIntervalClass('+P5')
                NamedIntervalClass('+M6')
                NamedIntervalClass('+m7')
                NamedIntervalClass('+M7')
                NamedIntervalClass('+P8')

        Returns interval set.
        """
        interval_set = IntervalSet.from_selection(selection)
        return class_(items=interval_set, item_class=item_class)


class IntervalSet(Set):
    """
    Interval set.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        prototype = (
            PitchClassSegment,
            PitchClassSet,
            PitchSegment,
            PitchSet,
        )
        if isinstance(items, prototype):
            items = list(items)
            pairs = enumerate.yield_pairs(items)
            items = [second - first for first, second in pairs]
        super().__init__(items=items, item_class=item_class)

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedInterval

    @property
    def _numbered_item_class(self):
        return NumberedInterval

    @property
    def _parent_item_class(self):
        return Interval

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Initializes interval set from component selection.

        ..  container:: example

            >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
            >>> staff_2 = abjad.Staff("c4. r8 g2")
            >>> selection = abjad.select((staff_1, staff_2))
            >>> intervals = abjad.IntervalSet.from_selection(selection)
            >>> for interval in sorted(intervals):
            ...     interval
            ...
            NamedInterval('-M6')
            NamedInterval('-P5')
            NamedInterval('-A4')
            NamedInterval('-M3')
            NamedInterval('-m3')
            NamedInterval('-M2')
            NamedInterval('+m2')
            NamedInterval('+m3')
            NamedInterval('+M3')
            NamedInterval('+P4')
            NamedInterval('+P5')
            NamedInterval('+m7')
            NamedInterval('+M7')
            NamedInterval('+P8')
            NamedInterval('+M9')
            NamedInterval('+A11')
            NamedInterval('+M13')

        Returns interval set.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        pairs = enumerate.yield_pairs(pitch_segment)
        intervals = [second - first for first, second in pairs]
        return class_(items=intervals, item_class=item_class)


class PitchClassSet(Set):
    """
    Pitch-class set.

    ..  container:: example

        Initializes numbered pitch-class set:

        >>> numbered_pitch_class_set = abjad.PitchClassSet(
        ...     items=[-2, -1.5, 6, 7, -1.5, 7],
        ...     item_class=abjad.NumberedPitchClass,
        ...     )
        >>> numbered_pitch_class_set
        PitchClassSet([6, 7, 10, 10.5])

    ..  container:: example

        Initializes named pitch-class set:

        >>> named_pitch_class_set = abjad.PitchClassSet(
        ...     items=['c', 'ef', 'bqs,', 'd'],
        ...     item_class=abjad.NamedPitchClass,
        ...     )
        >>> named_pitch_class_set
        PitchClassSet(['c', 'd', 'ef', 'bqs'])

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __contains__(self, argument):
        """
        Is true when pitch-class set contains ``argument``.

        ..  container:: example

            Initializes numbered pitch-class set:

            >>> set_ = abjad.PitchClassSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitchClass,
            ...     )
            >>> set_
            PitchClassSet([6, 7, 10, 10.5])

            >>> abjad.NamedPitch('fs') in set_
            True

            >>> abjad.NamedPitch('f') in set_
            False

            >>> 6 in set_
            True

            >>> 5 in set_
            False

        Returns true or false.
        """
        return super().__contains__(argument)

    def __hash__(self):
        """
        Hashes pitch-class set.

        Returns integer.
        """
        return super().__hash__()

    def __str__(self) -> str:
        """
        Gets string representation of pitch-class set.

        ..  container:: example

            Gets string of set sorted at initialization:

            >>> pc_set = abjad.PitchClassSet([6, 7, 10, 10.5])
            >>> str(pc_set)
            'PC{6, 7, 10, 10.5}'

        ..  container:: example

            Gets string of set not sorted at initialization:

            >>> pc_set = abjad.PitchClassSet([10.5, 10, 7, 6])
            >>> str(pc_set)
            'PC{6, 7, 10, 10.5}'

        """
        items = [str(_) for _ in sorted(self)]
        separator = " "
        if self.item_class is NumberedPitchClass:
            separator = ", "
        return f"PC{{{separator.join(items)}}}"

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedPitchClass

    @property
    def _numbered_item_class(self):
        return NumberedPitchClass

    @property
    def _parent_item_class(self):
        return PitchClass

    ### PRIVATE METHODS ###

    @staticmethod
    def _get_most_compact_ordering(candidates):
        widths = []
        for candidate in candidates:
            if candidate[0] < candidate[-1]:
                width = abs(candidate[-1] - candidate[0])
            else:
                width = abs(candidate[-1] + 12 - candidate[0])
            widths.append(width)
        minimum_width = min(widths)
        candidates_ = []
        for candidate, width in zip(candidates, widths):
            if width == minimum_width:
                candidates_.append(candidate)
        candidates = candidates_
        assert 1 <= len(candidates)
        if len(candidates) == 1:
            segment = candidates[0]
            segment = PitchClassSegment(items=segment, item_class=NumberedPitchClass)
            return segment
        for i in range(len(candidates[0]) - 1):
            widths = []
            for candidate in candidates:
                stop = i + 1
                if candidate[0] < candidate[stop]:
                    width = abs(candidate[stop] - candidate[0])
                else:
                    width = abs(candidate[stop] + 12 - candidate[0])
                widths.append(width)
            minimum_width = min(widths)
            candidates_ = []
            for candidate, width in zip(candidates, widths):
                if width == minimum_width:
                    candidates_.append(candidate)
            candidates = candidates_
            if len(candidates) == 1:
                segment = candidates[0]
                segment = PitchClassSegment(
                    items=segment, item_class=NumberedPitchClass
                )
                return segment
        candidates.sort(key=lambda x: x[0])
        segment = candidates[0]
        segment = PitchClassSegment(items=segment, item_class=NumberedPitchClass)
        return segment

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes pitch-class set from ``selection``.

        ..  container:: example

            >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
            >>> staff_2 = abjad.Staff("c4. r8 g2")
            >>> selection = abjad.select((staff_1, staff_2))
            >>> abjad.PitchClassSet.from_selection(selection)
            PitchClassSet(['c', 'd', 'fs', 'g', 'a', 'b'])

        Returns pitch-class set.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        return class_(items=pitch_segment, item_class=item_class)

    def get_normal_order(self):
        """
        Gets normal order.

        ..  container:: example

            Gets normal order of empty pitch-class set:

            >>> pc_set = abjad.PitchClassSet()
            >>> pc_set.get_normal_order()
            PitchClassSegment([])

        ..  container:: example

            Gets normal order:

            >>> pc_set = abjad.PitchClassSet([0, 1, 10, 11])
            >>> pc_set.get_normal_order()
            PitchClassSegment([10, 11, 0, 1])

        ..  container:: example

            Gets normal order:

            >>> pc_set = abjad.PitchClassSet([2, 8, 9])
            >>> pc_set.get_normal_order()
            PitchClassSegment([8, 9, 2])

        ..  container:: example

            Gets normal order of pitch-class set with degree of symmetry equal
            to 2:

            >>> pc_set = abjad.PitchClassSet([1, 2, 7, 8])
            >>> pc_set.get_normal_order()
            PitchClassSegment([1, 2, 7, 8])

        ..  container:: example

            Gets normal order of pitch-class set with degree of symmetry equal
            to 4:

            >>> pc_set = abjad.PitchClassSet([0, 3, 6, 9])
            >>> pc_set.get_normal_order()
            PitchClassSegment([0, 3, 6, 9])

        Returns pitch-class segment.
        """
        if not len(self):
            return PitchClassSegment(items=None, item_class=NumberedPitchClass)
        pitch_classes = list(self)
        pitch_classes.sort()
        candidates = []
        for i in range(self.cardinality):
            candidate = [NumberedPitch(_) for _ in pitch_classes]
            candidate = Sequence(candidate).rotate(n=-i)
            candidates.append(candidate)
        return self._get_most_compact_ordering(candidates)

    def get_prime_form(self, transposition_only=False):
        """
        Gets prime form.

        ..  container:: example

            Gets prime form of empty pitch-class set:

            >>> pc_set = abjad.PitchClassSet()
            >>> pc_set.get_prime_form()
            PitchClassSet([])

            >>> pc_set = abjad.PitchClassSet()
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([])

        ..  container:: example

            Gets prime form:

            >>> pc_set = abjad.PitchClassSet([0, 1, 10, 11])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 1, 2, 3])

            >>> pc_set = abjad.PitchClassSet([0, 1, 10, 11])
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([0, 1, 2, 3])

        ..  container:: example

            Gets prime form:

            >>> pc_set = abjad.PitchClassSet([2, 8, 9])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 1, 6])

            >>> pc_set = abjad.PitchClassSet([2, 8, 9])
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([0, 1, 6])

        ..  container:: example

            Gets prime form of pitch-class set with degree of symmetry equal to
            2:

            >>> pc_set = abjad.PitchClassSet([1, 2, 7, 8])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 1, 6, 7])

            >>> pc_set = abjad.PitchClassSet([1, 2, 7, 8])
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([0, 1, 6, 7])

        ..  container:: example

            Gets prime form of pitch-class set with degree of symmetry equal to
            4:

            >>> pc_set = abjad.PitchClassSet([0, 3, 6, 9])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 3, 6, 9])

            >>> pc_set = abjad.PitchClassSet([0, 3, 6, 9])
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([0, 3, 6, 9])

        ..  container:: example

            Gets prime form of pitch-class that is not inversion-equivalent:

            >>> pc_set = abjad.PitchClassSet([0, 4, 6, 7])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 1, 3, 7])

            >>> pc_set = abjad.PitchClassSet([0, 4, 6, 7])
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([0, 4, 6, 7])

        ..  container:: example

            Gets prime form of inversionally nonequivalent pitch-class set:

            >>> pc_set = abjad.PitchClassSet([0, 4, 7])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 3, 7])

            >>> pc_set = abjad.PitchClassSet([0, 4, 7])
            >>> pc_set.get_prime_form(transposition_only=True)
            PitchClassSet([0, 4, 7])

        ..  container:: example

            REGRESSION:

            >>> pc_set = abjad.PitchClassSet([0, 1, 2, 5, 8, 9])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 1, 2, 5, 6, 9])

            REGRESSION:

            >>> pc_set = abjad.PitchClassSet([0, 1, 2, 3, 6, 7])
            >>> pc_set.get_prime_form()
            PitchClassSet([0, 1, 2, 3, 6, 7])

        Returns new pitch-class set.
        """
        if not len(self):
            return copy.copy(self)
        normal_order = self.get_normal_order()
        if not transposition_only:
            normal_orders = [normal_order]
            inversion = self.invert()
            normal_order = inversion.get_normal_order()
            normal_orders.append(normal_order)
            normal_orders = [_._transpose_to_zero() for _ in normal_orders]
            assert len(normal_orders) == 2
            for left_pc, right_pc in zip(*normal_orders):
                if left_pc == right_pc:
                    continue
                if left_pc < right_pc:
                    normal_order = normal_orders[0]
                    break
                if right_pc < left_pc:
                    normal_order = normal_orders[-1]
                    break
        pcs = [_.number for _ in normal_order]
        first_pc = pcs[0]
        pcs = [pc - first_pc for pc in pcs]
        prime_form = type(self)(items=pcs, item_class=NumberedPitchClass)
        return prime_form

    def invert(self, axis=None):
        """
        Inverts pitch-class set.

        ..  container:: example

            >>> abjad.PitchClassSet(
            ...     [-2, -1.5, 6, 7, -1.5, 7],
            ...     ).invert()
            PitchClassSet([1.5, 2, 5, 6])

        Returns numbered pitch-class set.
        """
        return type(self)([pc.invert(axis=axis) for pc in self])

    def is_transposed_subset(self, pcset):
        """
        Is true when pitch-class set is transposed subset of ``pcset``.

        ..  container:: example

            >>> pitch_class_set_1 = abjad.PitchClassSet(
            ...     [-2, -1.5, 6, 7, -1.5, 7],
            ...     )
            >>> pitch_class_set_2 = abjad.PitchClassSet(
            ...     [-2, -1.5, 6, 7, -1.5, 7, 7.5, 8],
            ...     )

            >>> pitch_class_set_1.is_transposed_subset(pitch_class_set_2)
            True

        Returns true or false.
        """
        for n in range(12):
            if self.transpose(n).issubset(pcset):
                return True
        return False

    def is_transposed_superset(self, pcset):
        """
        Is true when pitch-class set is transposed superset of ``pcset``.

        ..  container:: example

            >>> pitch_class_set_1 = abjad.PitchClassSet(
            ...     [-2, -1.5, 6, 7, -1.5, 7],
            ...     )
            >>> pitch_class_set_2 = abjad.PitchClassSet(
            ...     [-2, -1.5, 6, 7, -1.5, 7, 7.5, 8],
            ...     )

            >>> pitch_class_set_2.is_transposed_superset(pitch_class_set_1)
            True

        Returns true or false.
        """
        for n in range(12):
            if self.transpose(n).issuperset(pcset):
                return True
        return False

    def multiply(self, n):
        """
        Multiplies pitch-class set by ``n``.

        ..  container:: example

            >>> abjad.PitchClassSet(
            ...     [-2, -1.5, 6, 7, -1.5, 7],
            ...     ).multiply(5)
            PitchClassSet([2, 4.5, 6, 11])

        Returns new pitch-class set.
        """
        items = (pitch_class.multiply(n) for pitch_class in self)
        return new(self, items=items)

    def order_by(self, segment):
        """
        Orders pitch-class set by pitch-class ``segment``.

        ..  container:: example

            >>> set_ = abjad.PitchClassSet(['c', 'e', 'b'])
            >>> segment = abjad.PitchClassSegment(['e', 'a', 'f'])
            >>> set_.order_by(segment)
            PitchClassSegment("b e c")

        Returns pitch-class segment.
        """
        if not len(self) == len(segment):
            raise ValueError("set and segment must be on equal length.")
        for pitch_classes in enumerate.yield_permutations(self):
            candidate = PitchClassSegment(pitch_classes)
            if candidate._is_equivalent_under_transposition(segment):
                return candidate
        raise ValueError(f"{self!s} can not order by {segment!s}.")

    def transpose(self, n=0):
        """
        Transposes all pitch-classes in pitch-class set by index ``n``.

        ..  container:: example

            >>> set_ = abjad.PitchClassSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitchClass,
            ...     )

            >>> for n in range(12):
            ...     print(n, set_.transpose(n))
            ...
            0 PC{6, 7, 10, 10.5}
            1 PC{7, 8, 11, 11.5}
            2 PC{0, 0.5, 8, 9}
            3 PC{1, 1.5, 9, 10}
            4 PC{2, 2.5, 10, 11}
            5 PC{0, 3, 3.5, 11}
            6 PC{0, 1, 4, 4.5}
            7 PC{1, 2, 5, 5.5}
            8 PC{2, 3, 6, 6.5}
            9 PC{3, 4, 7, 7.5}
            10 PC{4, 5, 8, 8.5}
            11 PC{5, 6, 9, 9.5}

        Returns new pitch-class set.
        """
        items = (pitch_class + n for pitch_class in self)
        return new(self, items=items)


class PitchSet(Set):
    r"""
    Pitch set.

    ..  container:: example

        Numbered pitch set:

        >>> set_ = abjad.PitchSet(
        ...     items=[-2, -1.5, 6, 7, -1.5, 7],
        ...     item_class=abjad.NumberedPitch,
        ...     )
        >>> set_
        PitchSet([-2, -1.5, 6, 7])

        >>> string = abjad.storage(set_)
        >>> print(string)
        abjad.PitchSet(
            [-2, -1.5, 6, 7]
            )

    ..  container:: example

        Named pitch set:

        >>> set_ = abjad.PitchSet(
        ...     ['bf,', 'aqs', "fs'", "g'", 'bqf', "g'"],
        ...     item_class=abjad.NamedPitch,
        ...     )
        >>> set_
        PitchSet(['bf,', 'aqs', 'bqf', "fs'", "g'"])

        >>> string = abjad.storage(set_)
        >>> print(string)
        abjad.PitchSet(
            ['bf,', 'aqs', 'bqf', "fs'", "g'"]
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when pitch set equals ``argument``.

        ..  container:: example

            >>> set_1 = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> set_2 = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> set_3 = abjad.PitchSet(
            ...     items=[11, 12, 12.5],
            ...     item_class=abjad.NumberedPitch,
            ...     )

            >>> set_1 == set_1
            True
            >>> set_1 == set_2
            True
            >>> set_1 == set_3
            False

            >>> set_2 == set_1
            True
            >>> set_2 == set_2
            True
            >>> set_2 == set_3
            False

            >>> set_3 == set_1
            False
            >>> set_3 == set_2
            False
            >>> set_3 == set_3
            True

        Return true or false.
        """
        return super().__eq__(argument)

    def __hash__(self):
        """
        Hashes pitch set.

        Returns number.
        """
        return super().__hash__()

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedPitch

    @property
    def _numbered_item_class(self):
        return NumberedPitch

    @property
    def _parent_item_class(self):
        return Pitch

    ### PRIVATE METHODS ###

    def _is_equivalent_under_transposition(self, argument):
        """
        True if pitch set is equivalent to ``argument`` under transposition.

        Returns true or false.
        """
        if not isinstance(argument, type(self)):
            return False
        if not len(self) == len(argument):
            return False
        difference = -(NamedPitch(argument[0], 4) - NamedPitch(self[0], 4))
        new_pitches = (x + difference for x in self)
        new_pitches = new(self, items=new_pitches)
        return argument == new_pitches

    def _sort_self(self):
        return sorted(PitchSegment(tuple(self)))

    ### PUBLIC PROPERTIES ###

    @property
    def duplicate_pitch_classes(self):
        """
        Gets duplicate pitch-classes in pitch set.

        ..  container:: example

            >>> set_ = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> set_.duplicate_pitch_classes
            PitchClassSet([])

            >>> set_ = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, 10.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> set_.duplicate_pitch_classes
            PitchClassSet([10.5])

        Returns pitch-class set.
        """
        pitch_classes = []
        duplicate_pitch_classes = []
        for pitch in self:
            pitch_class = NumberedPitchClass(pitch)
            if pitch_class in pitch_classes:
                duplicate_pitch_classes.append(pitch_class)
            pitch_classes.append(pitch_class)
        return PitchClassSet(duplicate_pitch_classes, item_class=NumberedPitchClass)

    @property
    def hertz(self):
        """
        Gets hertz of pitches in pitch segment.

        ..  container:: example

            >>> pitch_set = abjad.PitchSet('c e g b')
            >>> sorted(pitch_set.hertz)
            [130.81..., 164.81..., 195.99..., 246.94...]

        Returns set.
        """
        return set(_.hertz for _ in self)

    @property
    def is_pitch_class_unique(self):
        """
        Is true when pitch set is pitch-class-unique.

        ..  container:: example

            >>> set_ = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )

            >>> set_.is_pitch_class_unique
            True

        ..  container:: example

            >>> set_ = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, 10.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )

            >>> set_.is_pitch_class_unique
            False

        Returns true or false.
        """
        numbered_pitch_class_set = PitchClassSet(self, item_class=NumberedPitchClass)
        return len(self) == len(numbered_pitch_class_set)

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes pitch set from ``selection``.

        ..  container:: example

            >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
            >>> staff_2 = abjad.Staff("c4. r8 g2")
            >>> selection = abjad.select((staff_1, staff_2))
            >>> abjad.PitchSet.from_selection(selection)
            PitchSet(['c', 'g', 'b', "c'", "d'", "fs'", "a'"])

        Returns pitch set.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        return class_(items=pitch_segment, item_class=item_class)

    def invert(self, axis):
        """
        Inverts pitch set about ``axis``.

        Returns new pitch set.
        """
        items = (pitch.invert(axis) for pitch in self)
        return new(self, items=items)

    def issubset(self, argument):
        """
        Is true when pitch set is subset of ``argument``.

        ..  container:: example

            >>> set_1 = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> set_2 = abjad.PitchSet(
            ...     items=[-1.5, 6],
            ...     item_class=abjad.NumberedPitch,
            ...     )

            >>> set_1.issubset(set_2)
            False

            >>> set_2.issubset(set_1)
            True

        Returns true or false.
        """
        return super().issubset(argument)

    def issuperset(self, argument):
        """
        Is true when pitch set is superset of ``argument``.

        ..  container:: example

            >>> set_1 = abjad.PitchSet(
            ...     items=[-2, -1.5, 6, 7, -1.5, 7],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> set_2 = abjad.PitchSet(
            ...     items=[-1.5, 6],
            ...     item_class=abjad.NumberedPitch,
            ...     )

            >>> set_1.issuperset(set_2)
            False

            >>> set_2.issuperset(set_1)
            True

        Returns true or false.
        """
        return super().issubset(argument)

    def register(self, pitch_classes):
        """
        Registers ``pitch_classes`` by pitch set.

        ..  container:: example

            >>> pitch_set = abjad.PitchSet(
            ...     items=[10, 19, 20, 23, 24, 26, 27, 29, 30, 33, 37, 40],
            ...     item_class=abjad.NumberedPitch,
            ...     )
            >>> pitch_classes = [10, 0, 2, 6, 8, 7, 5, 3, 1, 9, 4, 11]
            >>> pitches = pitch_set.register(pitch_classes)
            >>> for pitch in pitches:
            ...     pitch
            NumberedPitch(10)
            NumberedPitch(24)
            NumberedPitch(26)
            NumberedPitch(30)
            NumberedPitch(20)
            NumberedPitch(19)
            NumberedPitch(29)
            NumberedPitch(27)
            NumberedPitch(37)
            NumberedPitch(33)
            NumberedPitch(40)
            NumberedPitch(23)

        Returns list of zero or more numbered pitches.
        """
        if isinstance(pitch_classes, collections.abc.Iterable):
            result = [
                [_ for _ in self if _.number % 12 == pc]
                for pc in [x % 12 for x in pitch_classes]
            ]
            result = Sequence(result).flatten(depth=-1)
        elif isinstance(pitch_classes, int):
            result = [p for p in pitch_classes if p % 12 == pitch_classes][0]
        else:
            raise TypeError("must be pitch-class or list of pitch-classes.")
        return result

    def transpose(self, n=0):
        """
        Transposes pitch set by index ``n``.

        Returns new pitch set.
        """
        items = (pitch.transpose(n=n) for pitch in self)
        return new(self, items=items)



from . import bundle as _bundle
from . import enums, overrides, storage
from . import tag as _tag
from .new import new


def remove_tags(string) -> str:
    """
    Removes all tags from ``string``.
    """
    lines = []
    for line in string.split("\n"):
        if "%!" not in line:
            lines.append(line)
            continue
        tag_start = line.find("%!")
        line = line[:tag_start]
        line = line.rstrip()
        if line:
            lines.append(line)
    string = "\n".join(lines)
    return string


class LilyPondFormatManager:
    """
    Manages LilyPond formatting logic.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "LilyPond formatting"

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return storage.StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    @staticmethod
    def _collect_indicators(component):
        wrappers = []
        for parent in component._get_parentage():
            wrappers_ = parent._get_indicators(unwrap=False)
            wrappers.extend(wrappers_)
        up_markup_wrappers = []
        down_markup_wrappers = []
        neutral_markup_wrappers = []
        context_wrappers = []
        noncontext_wrappers = []
        # classify wrappers attached to component
        for wrapper in wrappers:
            # skip nonprinting indicators like annotation
            indicator = wrapper.indicator
            if not hasattr(indicator, "_get_lilypond_format") and not hasattr(
                indicator, "_get_lilypond_format_bundle"
            ):
                continue
            elif wrapper.annotation is not None:
                continue
            # skip comments and commands unless attached directly to us
            elif (
                wrapper.context is None
                and hasattr(wrapper.indicator, "_format_leaf_children")
                and not getattr(wrapper.indicator, "_format_leaf_children")
                and wrapper.component is not component
            ):
                continue
            # store markup wrappers
            elif wrapper.indicator.__class__.__name__ == "Markup":
                if wrapper.indicator.direction is enums.Up:
                    up_markup_wrappers.append(wrapper)
                elif wrapper.indicator.direction is enums.Down:
                    down_markup_wrappers.append(wrapper)
                elif wrapper.indicator.direction in (enums.Center, None):
                    neutral_markup_wrappers.append(wrapper)
            # store context wrappers
            elif wrapper.context is not None:
                if wrapper.annotation is None and wrapper.component is component:
                    context_wrappers.append(wrapper)
            # store noncontext wrappers
            else:
                noncontext_wrappers.append(wrapper)
        indicators = (
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
            context_wrappers,
            noncontext_wrappers,
        )
        return indicators

    @staticmethod
    def _populate_context_setting_format_contributions(component, bundle):
        result = []
        if hasattr(component, "_lilypond_type"):
            strings = overrides.setting(component)._format_in_with_block()
            result.extend(strings)
        else:
            strings = overrides.setting(component)._format_inline()
            result.extend(strings)
        result.sort()
        bundle.context_settings.extend(result)

    @staticmethod
    def _populate_context_wrapper_format_contributions(
        component, bundle, context_wrappers
    ):
        for wrapper in context_wrappers:
            format_pieces = wrapper._get_format_pieces()
            if isinstance(format_pieces, type(bundle)):
                bundle.update(format_pieces)
            else:
                format_slot = wrapper.indicator._format_slot
                bundle.get(format_slot).indicators.extend(format_pieces)

    @staticmethod
    def _populate_grob_override_format_contributions(component, bundle):
        result = []
        once = hasattr(component, "_written_duration")
        grob = overrides.override(component)
        contributions = grob._list_format_contributions("override", once=once)
        for string in result[:]:
            if "NoteHead" in string and "pitch" in string:
                contributions.remove(string)
        try:
            written_pitch = component.written_pitch
            arrow = written_pitch.arrow
        except AttributeError:
            arrow = None
        if arrow in (enums.Up, enums.Down):
            contributions_ = written_pitch._list_format_contributions()
            contributions.extend(contributions_)
        bundle.grob_overrides.extend(contributions)

    @staticmethod
    def _populate_grob_revert_format_contributions(component, bundle):
        if not hasattr(component, "_written_duration"):
            contributions = overrides.override(component)._list_format_contributions(
                "revert"
            )
            bundle.grob_reverts.extend(contributions)

    @staticmethod
    def _populate_indicator_format_contributions(component, bundle):
        (
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
            context_wrappers,
            noncontext_wrappers,
        ) = LilyPondFormatManager._collect_indicators(component)
        LilyPondFormatManager._populate_markup_format_contributions(
            component,
            bundle,
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
        )
        LilyPondFormatManager._populate_context_wrapper_format_contributions(
            component, bundle, context_wrappers
        )
        LilyPondFormatManager._populate_noncontext_wrapper_format_contributions(
            component, bundle, noncontext_wrappers
        )

    @staticmethod
    def _populate_markup_format_contributions(
        component,
        bundle,
        up_markup_wrappers,
        down_markup_wrappers,
        neutral_markup_wrappers,
    ):
        for wrappers in (
            up_markup_wrappers,
            down_markup_wrappers,
            neutral_markup_wrappers,
        ):
            for wrapper in wrappers:
                if wrapper.indicator.direction is None:
                    markup = new(wrapper.indicator, direction="-")
                else:
                    markup = wrapper.indicator
                format_pieces = markup._get_format_pieces()
                format_pieces = _tag.double_tag(
                    format_pieces, wrapper.tag, deactivate=wrapper.deactivate
                )
                bundle.after.markup.extend(format_pieces)

    @staticmethod
    def _populate_noncontext_wrapper_format_contributions(
        component, bundle, noncontext_wrappers
    ):
        for wrapper in noncontext_wrappers:
            indicator = wrapper.indicator
            if hasattr(indicator, "_get_lilypond_format_bundle"):
                bundle_ = indicator._get_lilypond_format_bundle()
                if wrapper.tag:
                    bundle_.tag_format_contributions(
                        wrapper.tag, deactivate=wrapper.deactivate
                    )
                if bundle_ is not None:
                    bundle.update(bundle_)

    @staticmethod
    def _report_leaf_format_contributions(leaf):
        bundle = LilyPondFormatManager.bundle_format_contributions(leaf)
        report = ""
        report += 'slot "absolute before":\n'
        packet = leaf._format_absolute_before_slot(bundle)
        report += leaf._process_contribution_packet(packet)
        report += 'slot "before":\n'
        packet = leaf._format_before_slot(bundle)
        report += leaf._process_contribution_packet(packet)
        report += 'slot "opening":\n'
        packet = leaf._format_opening_slot(bundle)
        report += leaf._process_contribution_packet(packet)
        report += 'slot "contents slot":\n'
        report += _bundle.LilyPondFormatBundle.indent + "leaf body:\n"
        string = leaf._format_contents_slot(bundle)[0][1][0]
        report += (2 * _bundle.LilyPondFormatBundle.indent) + string + "\n"
        report += 'slot "closing":\n'
        packet = leaf._format_closing_slot(bundle)
        report += leaf._process_contribution_packet(packet)
        report += 'slot "after":\n'
        packet = leaf._format_after_slot(bundle)
        report += leaf._process_contribution_packet(packet)
        report += 'slot "absolute after":\n'
        packet = leaf._format_absolute_after_slot(bundle)
        report += leaf._process_contribution_packet(packet)
        while report[-1] == "\n":
            report = report[:-1]
        return report

    ### PUBLIC METHODS ###

    @staticmethod
    def bundle_format_contributions(component) -> "_bundle.LilyPondFormatBundle":
        """
        Gets all format contributions for ``component``.
        """
        bundle = _bundle.LilyPondFormatBundle()
        LilyPondFormatManager._populate_indicator_format_contributions(
            component, bundle
        )
        LilyPondFormatManager._populate_context_setting_format_contributions(
            component, bundle
        )
        LilyPondFormatManager._populate_grob_override_format_contributions(
            component, bundle
        )
        LilyPondFormatManager._populate_grob_revert_format_contributions(
            component, bundle
        )
        bundle.sort_overrides()
        return bundle

    @staticmethod
    def left_shift_tags(text, realign=None) -> str:
        """
        Left shifts tags in ``strings`` and realigns to column ``realign``.
        """
        strings = text.split("\n")
        strings_ = []
        for string in strings:
            if "%@% " not in string or "%!" not in string:
                strings_.append(string)
                continue
            if not string.startswith(4 * " "):
                strings_.append(string)
                continue
            string_ = string[4:]
            tag_start = string_.find("%!")
            string_ = list(string_)
            string_[tag_start:tag_start] = _bundle.LilyPondFormatBundle.indent
            string_ = "".join(string_)
            strings_.append(string_)
        text = "\n".join(strings_)
        return text



from ..markups import Markup
from ..storage import StorageFormatManager


class StringContactPoint:
    """
    String contact point.

    ..  container:: example

        Sul ponticello:

        >>> indicator = abjad.StringContactPoint('sul ponticello')
        >>> string = abjad.storage(indicator)
        >>> print(string)
        abjad.StringContactPoint(
            contact_point='sul ponticello',
            )

    ..  container:: example

        Sul tasto:

        >>> indicator = abjad.StringContactPoint('sul tasto')
        >>> string = abjad.storage(indicator)
        >>> print(string)
        abjad.StringContactPoint(
            contact_point='sul tasto',
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_contact_point",)

    _contact_point_abbreviations = {
        "dietro ponticello": "d.p.",
        "molto sul ponticello": "m.s.p",
        "molto sul tasto": "m.s.t.",
        "ordinario": "ord.",
        "pizzicato": "pizz.",
        "ponticello": "p.",
        "sul ponticello": "s.p.",
        "sul tasto": "s.t.",
    }

    _contact_points = (
        "dietro ponticello",
        "molto sul ponticello",
        "molto sul tasto",
        "ordinario",
        "pizzicato",
        "ponticello",
        "sul ponticello",
        "sul tasto",
    )

    _parameter = "SCP"

    _persistent = True

    ### INITIALIZER ###

    def __init__(self, contact_point: str = "ordinario") -> None:
        contact_point = str(contact_point)
        assert contact_point in self._contact_points
        self._contact_point = contact_point

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PUBLIC PROPERTIES ###

    @property
    def contact_point(self) -> str:
        """
        Gets contact point of string contact point.

        ..  container:: example

            Sul ponticello:

            >>> indicator = abjad.StringContactPoint('sul ponticello')
            >>> indicator.contact_point
            'sul ponticello'

        ..  container:: example

            Sul tasto:

            >>> indicator = abjad.StringContactPoint('sul tasto')
            >>> indicator.contact_point
            'sul tasto'

        Set to known string.
        """
        return self._contact_point

    @property
    def markup(self) -> Markup:
        r"""
        Gets markup of string contact point.

        ..  container:: example

            Sul ponticello:

            >>> indicator = abjad.StringContactPoint('sul ponticello')
            >>> abjad.show(indicator.markup) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(indicator.markup)
                >>> print(string)
                \markup {
                    \caps
                        S.P.
                    }

        ..  container:: example

            Sul tasto:

            >>> indicator = abjad.StringContactPoint('sul tasto')
            >>> abjad.show(indicator.markup) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(indicator.markup)
                >>> print(string)
                \markup {
                    \caps
                        S.T.
                    }

        """
        string = self._contact_point_abbreviations[self.contact_point]
        string = rf"\caps {string.title()}"
        markup = Markup(string)
        return markup

    @property
    def parameter(self) -> str:
        """
        Returns ``'SCP'``.

        ..  container:: example

            >>> abjad.StringContactPoint('sul tasto').parameter
            'SCP'

        """
        return self._parameter

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.StringContactPoint('sul tasto').persistent
            True

        """
        return self._persistent

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on string contact point.
        """
        pass



import typing

from ..bundle import LilyPondFormatBundle
from ..overrides import TweakInterface
from ..pitch.intervals import NamedInterval
from ..pitch.pitches import NamedPitch
from ..storage import StorageFormatManager


class StartTrillSpan:
    r"""
    LilyPond ``\startTrillSpan`` command.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> start_trill_span = abjad.StartTrillSpan()
        >>> abjad.tweak(start_trill_span).color = "#blue"
        >>> abjad.attach(start_trill_span, staff[0])
        >>> stop_trill_span = abjad.StopTrillSpan()
        >>> abjad.attach(stop_trill_span, staff[-1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #blue
                \startTrillSpan
                d'4
                e'4
                f'4
                \stopTrillSpan
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_interval", "_pitch", "_tweaks")

    _context = "Voice"

    _parameter = "TRILL"

    _persistent = True

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        interval: typing.Union[str, NamedInterval] = None,
        pitch: typing.Union[str, NamedPitch] = None,
        tweaks: TweakInterface = None,
    ) -> None:
        if interval is not None:
            interval = NamedInterval(interval)
        self._interval = interval
        if pitch is not None:
            pitch = NamedPitch(pitch)
        self._pitch = pitch
        if tweaks is not None:
            assert isinstance(tweaks, TweakInterface), repr(tweaks)
        self._tweaks = TweakInterface.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            bundle.after.spanner_starts.extend(tweaks)
        string = r"\startTrillSpan"
        if self.interval or self.pitch:
            bundle.opening.spanners.append(r"\pitchedTrill")
            if self.pitch:
                pitch = self.pitch
            else:
                pitch = component.written_pitch + self.interval
            string = string + f" {pitch!s}"
        bundle.after.spanner_starts.append(string)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Returns (historically conventional) context ``'Voice'``.

        ..  container:: example

            >>> abjad.StartTrillSpan().context
            'Voice'

        Class constant.

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def interval(self) -> typing.Optional[NamedInterval]:
        r"""
        Gets interval.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> start_trill_span = abjad.StartTrillSpan(interval='M2')
            >>> abjad.tweak(start_trill_span).color = "#blue"
            >>> abjad.attach(start_trill_span, staff[0])
            >>> stop_trill_span = abjad.StopTrillSpan()
            >>> abjad.attach(stop_trill_span, staff[-1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \pitchedTrill
                    c'4
                    - \tweak color #blue
                    \startTrillSpan d'
                    d'4
                    e'4
                    f'4
                    \stopTrillSpan
                }

        """
        return self._interval

    @property
    def parameter(self) -> str:
        """
        Returns ``'TRILL'``.

        ..  container:: example

            >>> abjad.StartTrillSpan().parameter
            'TRILL'

        Class constant.
        """
        return self._parameter

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.StartTrillSpan().persistent
            True

        Class constant.
        """
        return self._persistent

    @property
    def pitch(self) -> typing.Optional[NamedPitch]:
        r"""
        Gets pitch.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> start_trill_span = abjad.StartTrillSpan(pitch='C#4')
            >>> abjad.tweak(start_trill_span).color = "#blue"
            >>> abjad.attach(start_trill_span, staff[0])
            >>> stop_trill_span = abjad.StopTrillSpan()
            >>> abjad.attach(stop_trill_span, staff[-1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \pitchedTrill
                    c'4
                    - \tweak color #blue
                    \startTrillSpan cs'
                    d'4
                    e'4
                    f'4
                    \stopTrillSpan
                }

        """
        return self._pitch

    @property
    def spanner_start(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.StartTrillSpan().spanner_start
            True

        """
        return True

    @property
    def tweaks(self) -> typing.Optional[TweakInterface]:
        r"""
        Gets tweaks

        ..  container:: example

            REGRESSION. Tweaks survive copy:

            >>> import copy
            >>> start_trill_span = abjad.StartTrillSpan()
            >>> abjad.tweak(start_trill_span).color = "#blue"
            >>> string = abjad.storage(start_trill_span)
            >>> print(string)
            abjad.StartTrillSpan(
                tweaks=TweakInterface(('_literal', None), ('color', '#blue')),
                )

            >>> start_trill_span_2 = copy.copy(start_trill_span)
            >>> string = abjad.storage(start_trill_span_2)
            >>> print(string)
            abjad.StartTrillSpan(
                tweaks=TweakInterface(('_literal', None), ('color', '#blue')),
                )

        """
        return self._tweaks
